"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var constants_1 = require("../constants");
var errors_1 = tslib_1.__importDefault(require("../errors"));
var utils_1 = require("../utils");
var FieldSchemaModel = /** @class */ (function () {
    function FieldSchemaModel(table, name, schema, cascadeAsDefault) {
        if (cascadeAsDefault === void 0) { cascadeAsDefault = false; }
        this.table = utils_1.ensureParamObject("table", table);
        utils_1.ensureParamString("name", name);
        utils_1.ensureParamObject("schema", schema);
        this.type = schema.type || constants_1.TYPE_ATTR;
        this.name = utils_1.optionalParamString("schema.fieldName", schema.fieldName, name);
        this.propName = utils_1.optionalParamString("schema.propName", schema.propName, name);
        this.references = utils_1.optionalParamString("schema.references", schema.references);
        this.isPrimaryKey = schema.pk === true || schema.type === constants_1.TYPE_PK;
        this.isForeignKey = !!this.references;
        this.isStamp = schema.stamp === true || schema.type === constants_1.TYPE_MODIFIED;
        this._valueFactory = schema.value !== undefined
            ? utils_1.ensureParamFunction("schema.value", schema.value).bind(this)
            : undefined;
        if (this.isPrimaryKey || this.isForeignKey) {
            if (this.isForeignKey)
                this.relationName = utils_1.optionalParamString("schema.relationName", schema.relationName);
            this.cascade = schema.cascade === undefined ? cascadeAsDefault : schema.cascade === true;
            this.unique = schema.unique === true;
            // not null is default true, for PK's and FK's
            this.notNull = schema.notNull === undefined ? true : schema.notNull === true;
        }
        else {
            this.cascade = false;
            this.unique = false;
            this.notNull = schema.notNull === true;
        }
    }
    Object.defineProperty(FieldSchemaModel.prototype, "refTable", {
        get: function () { return this._refTable; },
        enumerable: true,
        configurable: true
    });
    FieldSchemaModel.prototype.connect = function (schemas) {
        if (this.references) {
            this._refTable = schemas[this.references];
            if (!this._refTable)
                throw new Error(errors_1.default.fkInvalidReference(this.table.name, this.name, this.references));
        }
    };
    FieldSchemaModel.prototype.getValue = function (data, record) {
        return this._valueFactory ? this._valueFactory(data, {
            record: record,
            schema: this
        }) : data[this.name];
    };
    FieldSchemaModel.prototype.getRecordValue = function (record) {
        return this.getValue(record.value, record);
    };
    return FieldSchemaModel;
}());
exports.default = FieldSchemaModel;
