import * as tslib_1 from "tslib";
import { ALL } from "../constants";
import errors from "../errors";
import { ensureArray, ensureParamObject, isEqual, toObject } from "../utils";
var defaultOptions = {
    cascadeAsDefault: false
};
var getMappedFunction = function (map, key, defaultFn) {
    if (!map)
        return defaultFn;
    if (typeof map === "function")
        return map;
    else if (map[key])
        return map[key];
    else if (map[ALL])
        return map[ALL];
    return defaultFn;
};
var Database = /** @class */ (function () {
    function Database(schema, factory, options) {
        var _this = this;
        this.getRecordNormalizer = function (schemaName) {
            return getMappedFunction(_this.options.onNormalize, schemaName);
        };
        this.getPkGenerator = function (schemaName) {
            return getMappedFunction(_this.options.onGeneratePK, schemaName);
        };
        this.getRecordComparer = function (schemaName) {
            return getMappedFunction(_this.options.onRecordCompare, schemaName, isEqual);
        };
        this.getRecordMerger = function (schemaName) {
            return getMappedFunction(_this.options.onRecordMerge, schemaName);
        };
        this.schema = ensureParamObject("schema", schema);
        this.factory = ensureParamObject("factory", factory);
        this.options = tslib_1.__assign({}, defaultOptions, options);
        this.tables = Object.keys(schema).map(function (tableName) {
            return _this.factory.newTableSchema(_this, tableName, schema[tableName]);
        });
        this.tableMap = toObject(this.tables, function (table) { return table.name; });
        this.tables.forEach(function (table) { return table.connect(_this.tableMap); });
    }
    Database.prototype.combineReducers = function () {
        var _this = this;
        var reducers = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            reducers[_i] = arguments[_i];
        }
        return function (state, action) {
            if (state === void 0) { state = {}; }
            return _this.reduce(state, action, reducers);
        };
    };
    Database.prototype.reduce = function (state, action, reducers) {
        var _this = this;
        if (state === void 0) { state = {}; }
        var args = [];
        for (var _i = 3; _i < arguments.length; _i++) {
            args[_i - 3] = arguments[_i];
        }
        var session = this.createSession(state);
        ensureArray(reducers).forEach(function (reducer) {
            return reducer.apply(_this, [session.tables, action].concat(args));
        });
        return session.commit();
    };
    Database.prototype.createSession = function (state, options) {
        if (state === void 0) { state = {}; }
        return this.factory.newSession(this, state, tslib_1.__assign({ readOnly: false }, options));
    };
    Database.prototype.getTableSchema = function (name) {
        return this.tableMap[name];
    };
    Database.prototype.selectTables = function (state) {
        var _this = this;
        ensureParamObject("state", state);
        var tableSchemas = Object.keys(state)
            .filter(function (tableName) { return _this.tableMap[tableName]; })
            .map(function (tableName) { return _this.tableMap[tableName]; });
        var session = this.createSession(state, {
            readOnly: true,
            tableSchemas: tableSchemas
        });
        return session.tables;
    };
    Database.prototype.selectTable = function (tableState, schemaName) {
        var _a;
        ensureParamObject("tableState", tableState);
        var _b = tableState.name, name = _b === void 0 ? schemaName : _b;
        if (!name)
            throw new Error(errors.unknownTableState());
        return this.selectTables((_a = {},
            _a[name] = tableState,
            _a))[name];
    };
    return Database;
}());
export default Database;
